import os
import platform
import sys
from typing import Mapping, Optional, Callable, NamedTuple, IO, AnyStr

from ._windows import WindowsConsoleFeatures

WINDOWS = platform.system() == 'Windows'
_windows_console_features: Optional["WindowsConsoleFeatures"] = None


def get_windows_console_features() -> "WindowsConsoleFeatures":  # pragma: no cover
    global _windows_console_features
    if _windows_console_features is not None:
        return _windows_console_features
    from ._windows import get_windows_console_features

    _windows_console_features = get_windows_console_features()
    return _windows_console_features


class ConsoleDimensions(NamedTuple):
    """Size of the terminal."""

    width: int
    """The width of the console in 'cells'."""
    height: int
    """The height of the console in lines."""


class SimpleConsole:
    """This is port from `rich`. It is mostly a subset of a rich.console.Console"""

    _environ: Mapping[str, str] = os.environ

    def __init__(self, file: Optional[IO[str]] = None):
        self._file = file
        self._buffer = []

    @property
    def file(self) -> IO[str]:
        """Get the file object to write to."""
        file = self._file or sys.stdout
        return file

    @property
    def is_terminal(self) -> bool:
        """Check if the console is writing to a terminal.

        Returns:
            bool: True if the console writing to a device capable of
            understanding terminal codes, otherwise False.
        """
        isatty: Optional[Callable[[], bool]] = getattr(self.file, "isatty", None)
        try:
            return False if isatty is None else isatty()
        except ValueError:
            # in some situation (at the end of a pytest run for example) isatty() can raise
            # ValueError: I/O operation on closed file
            # return False because we aren't in a terminal anymore
            return False

    @property
    def is_dumb_terminal(self) -> bool:
        """Detect dumb terminal.

        Returns:
            bool: True if writing to a dumb terminal, otherwise False.

        """
        _term = self._environ.get("TERM", "")
        is_dumb = _term.lower() in ("dumb", "unknown")
        return self.is_terminal and is_dumb

    @property
    def legacy_windows(self) -> bool:
        return WINDOWS and not get_windows_console_features().vt

    @property
    def size(self) -> ConsoleDimensions:
        """Get the size of the console.

        Returns:
            ConsoleDimensions: A named tuple containing the dimensions.
        """
        if self.is_dumb_terminal:
            return ConsoleDimensions(80, 25)

        width: Optional[int] = None
        height: Optional[int] = None

        if WINDOWS:  # pragma: no cover
            try:
                width, height = os.get_terminal_size()
            except OSError:  # Probably not a terminal
                pass
        else:
            try:
                width, height = os.get_terminal_size(sys.__stdin__.fileno())
            except (AttributeError, ValueError, OSError):
                try:
                    width, height = os.get_terminal_size(sys.__stdout__.fileno())
                except (AttributeError, ValueError, OSError):
                    pass

        columns = self._environ.get("COLUMNS")
        if columns is not None and columns.isdigit():
            width = int(columns)
        lines = self._environ.get("LINES")
        if lines is not None and lines.isdigit():
            height = int(lines)

        # get_terminal_size can report 0, 0 if run from pseudo-terminal
        width = width or 80
        height = height or 25
        return ConsoleDimensions(
            width - self.legacy_windows,
            height
        )

    def write(self, s: AnyStr):
        self._buffer += s

    def flush(self):
        self.file.write(''.join(self._buffer))
        self._buffer = []
        self.file.flush()



import functools
import sys


def clamp(minimum, v, maximum):
    """Clamp a value within the range"""
    return max(minimum, min(v, maximum))


if sys.version_info.major < 3:
    raise RuntimeError

# Python<3.9 doesn't have `cache`. Shim with `lru_cache`
cache = functools.cache if sys.version_info.minor >= 9 else functools.lru_cache



# ##############################################
#
# This module is are mostly borrowed from `rich`
#
# ##############################################

# `rich` is an excellent, feature-rich and well-maintained library. But here are some reasons of
#   extracting from `rich` (apart from pursuing 0-dep)
#
#   1. To achieve simple terminal drawing, we don't need most of the features from `rich`.
#   2. We don't want the dependencies brought by `rich` (although they are not a lot)
#   3. We want to be able to backport to python 2.7



from enum import IntEnum
from typing import Dict, Callable


class ControlType(IntEnum):
    """Non-printable control codes which typically translate to ANSI codes."""

    BELL = 1
    CARRIAGE_RETURN = 2
    HOME = 3
    CLEAR = 4
    SHOW_CURSOR = 5
    HIDE_CURSOR = 6
    ENABLE_ALT_SCREEN = 7
    DISABLE_ALT_SCREEN = 8
    CURSOR_UP = 9
    CURSOR_DOWN = 10
    CURSOR_FORWARD = 11
    CURSOR_BACKWARD = 12
    CURSOR_MOVE_TO_COLUMN = 13
    CURSOR_MOVE_TO = 14
    ERASE_IN_LINE = 15


CONTROL_CODES_FORMAT: Dict[int, Callable[..., str]] = {
    ControlType.BELL: lambda: "\x07",
    ControlType.CARRIAGE_RETURN: lambda: "\r",
    ControlType.HOME: lambda: "\x1b[H",
    ControlType.CLEAR: lambda: "\x1b[2J",
    ControlType.ENABLE_ALT_SCREEN: lambda: "\x1b[?1049h",
    ControlType.DISABLE_ALT_SCREEN: lambda: "\x1b[?1049l",
    ControlType.SHOW_CURSOR: lambda: "\x1b[?25h",
    ControlType.HIDE_CURSOR: lambda: "\x1b[?25l",
    ControlType.CURSOR_UP: lambda param: f"\x1b[{param}A",
    ControlType.CURSOR_DOWN: lambda param: f"\x1b[{param}B",
    ControlType.CURSOR_FORWARD: lambda param: f"\x1b[{param}C",
    ControlType.CURSOR_BACKWARD: lambda param: f"\x1b[{param}D",
    ControlType.CURSOR_MOVE_TO_COLUMN: lambda param: f"\x1b[{param+1}G",
    ControlType.ERASE_IN_LINE: lambda param: f"\x1b[{param}K",
    ControlType.CURSOR_MOVE_TO: lambda x, y: f"\x1b[{y+1};{x+1}H",
}



import asyncio
import heapq
import platform
import time
from typing import NamedTuple, List, Optional, Dict, Union, Tuple

from ._rich.control import ControlType, CONTROL_CODES_FORMAT
from ._rich.simple_console import SimpleConsole
from .utils import clamp, cache
from .screen import ScreenPoint

WINDOWS = platform.system() == 'Windows'


async def sleep(second: float):
    if second <= 0:
        return
    if second > 0.1:
        return await asyncio.sleep(second)
    start = time.perf_counter()
    while time.perf_counter() - start < second:
        continue


class Render(SimpleConsole):

    @property
    @cache
    def screen_size(self):
        return self.size

    def move_cursor_home(self):
        char_control = CONTROL_CODES_FORMAT[ControlType.HOME]()
        self.write(char_control)
        self.flush()

    def clear(self):
        char_control = CONTROL_CODES_FORMAT[ControlType.CLEAR]()
        self.write(char_control)
        self.flush()

    def string_at(self, p: Union[Tuple, ScreenPoint], s: str) -> str:
        """Use the terminal control character to position a string"""

        if not isinstance(p, ScreenPoint):
            p = ScreenPoint(*p)

        # if p.x > self.screen_size.width - 1:
        #     return
        # if p.y > self.screen_size.height - 2:
        #     return
        # if p.x < 0:
        #     return
        # if p.y < 0:
        #     return
        # x = int(p.x)
        # y = int(p.y)

        x = int(clamp(0, p.x, self.screen_size.width - 1))
        y = int(clamp(0, p.y, self.screen_size.height - 2))

        char_control = CONTROL_CODES_FORMAT[ControlType.CURSOR_MOVE_TO](x, y)
        return char_control + s

    def draw_string_at(self, p: Union[Tuple, ScreenPoint], s: str, flush=True):
        self.write(self.string_at(p, s))

        if flush:
            self.flush()


class AnimationRender(Render):
    CLEAR_CHAR = ' '

    class TimedDrawStruct(NamedTuple):
        frame_idx: int
        point: ScreenPoint
        char: str

    class TimedDrawFullFrame(NamedTuple):
        frame_idx: int
        buffer: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # This will be maintained sorted (priority queue)
        self._scheduled: List[AnimationRender.TimedDrawStruct] = []
        self._evt_stop = asyncio.Event()

    def schedule_draw(
            self,
            frame: int, p: ScreenPoint, s: str,
            clean_after: Optional[int] = None
    ):
        heapq.heappush(self._scheduled, AnimationRender.TimedDrawStruct(frame, p, s))

        if clean_after is not None:
            self.schedule_draw(frame + clean_after, p, self.CLEAR_CHAR, clean_after=None)

    def _process_frames(self, frames: List[TimedDrawStruct]) -> List[TimedDrawFullFrame]:
        """
        Return a time-indexed full frame. The potential issue of this implementation is that,
        The full frame string will be fixed during such `process`-time, so if terminal got resized this,
        the drawing will be messed up.

        Assuming `frames` are already sorted (by time).
        """
        if len(frames) == 0:
            return []

        frame_list: List[AnimationRender.TimedDrawFullFrame] = []

        def _insert(chuck: List[AnimationRender.TimedDrawStruct]):
            if not chuck:
                return

            frame_idx = chuck[0].frame_idx
            frame: Dict[ScreenPoint, str] = {}

            for tds in chuck:
                if tds.frame_idx != frame_idx:
                    raise RuntimeError('All the `TimedDrawStruct`s must have the same frame index')

                same_pos = frame.get(tds.point)
                if tds.char == self.CLEAR_CHAR and same_pos and same_pos != self.CLEAR_CHAR:
                    # If the (x, y) is already scheduled with a non-clear-char, don't put the clear-char here.
                    continue
                frame[tds.point] = tds.char

            frame_str = ''.join([self.string_at(p, frame[p]) for p in frame])
            frame_list.append(AnimationRender.TimedDrawFullFrame(frame_idx, frame_str))

        # Chunk all the frames in O(n) time
        slow, fast = 0, 1
        while True:
            if fast == len(frames):
                _insert(frames[slow:fast])
                break
            if frames[fast].frame_idx != frames[slow].frame_idx:
                _insert(frames[slow:fast])
                slow = fast
            fast += 1

        return frame_list

    async def render_frames(self, frame_interval_s=0.005, min_sleep_delay=0):

        # NOTED: We assume all frames are already scheduled i.e. `self._scheduled` is fixed

        if len(self._scheduled) == 0:
            return

        full_frame_list = self._process_frames(
            # Pop all the items (sorted)
            heapq.nsmallest(len(self._scheduled), self._scheduled)
        )

        current_frame = 0
        for frame in full_frame_list:
            empty_frames = frame.frame_idx - current_frame

            if empty_frames > 0:
                sleep_time = frame_interval_s * empty_frames
                if sleep_time > min_sleep_delay:
                    await sleep(sleep_time)

            self.write(frame.buffer)
            self.flush()

            current_frame = frame.frame_idx



import asyncio
import random
import sys

from .fun import all_art
from .brush import BrushWipe
from .path import PathZigZag, PathRectEdge
from .renderer import AnimationRender


def main(
        frame_interval_s=0.003,
        clean_after=None,
        min_frame_delay=0
):
    if clean_after is None:
        clean_after = 1

    r = AnimationRender()

    # You might customize the wipe brush
    bw = BrushWipe()

    common_brush_cfg = dict(
        size=bw.width,
        brush_deformation_factor=bw.deformation_factor,
        max_x=r.screen_size.width,
        max_y=r.screen_size.height,
    )

    # You might create your own wipe path
    path_points = PathZigZag(**common_brush_cfg).get_points_list()
    path_points += PathRectEdge(path_points[-1], **common_brush_cfg).get_points_list()

    for idx, pp in enumerate(path_points):
        for bwp in bw.get_points(*pp.coord, pp.angle):
            r.schedule_draw(
                frame=idx,
                p=bwp.coord,
                s=bwp.char,
                clean_after=clean_after
            )

    try:
        asyncio.new_event_loop().run_until_complete(
            r.render_frames(frame_interval_s, min_frame_delay)
        )
    except KeyboardInterrupt:
        pass

    r.clear()
    r.move_cursor_home()


def cli(*args):
    import argparse
    from argparse import RawTextHelpFormatter

    text = """
    wipe-clean - Clean your terminal in a ritual way
    """

    default = {
        'f': 0.003,
        'c': 1,
        'm': 0
    }

    parser = argparse.ArgumentParser(description=random.choice(all_art) + text, formatter_class=RawTextHelpFormatter)
    parser.add_argument('-f', '--frame-interval', default=default['f'],
                        type=float, help='Frame interval (in second)')
    parser.add_argument('-c', '--clean-after', default=default['c'],
                        type=float, help='Clean drawn after number of frame(s)')
    parser.add_argument('-m', '--min-frame-delay', default=default['m'],
                        type=float, help='Minimum frame delay (in second). A delay will only be will be'
                                         ' scheduled when frame interval is larger than this value.'
                                         ' This may help solve the inaccurate sleep on Windows.')

    parsed_args = parser.parse_args(args)
    main(
        frame_interval_s=parsed_args.frame_interval,
        clean_after=parsed_args.clean_after,
        min_frame_delay=parsed_args.min_frame_delay
    )


def _outer_cli():
    sys.exit(
        cli(*sys.argv[1:])
    )


if __name__ == '__main__':
    _outer_cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wipe_clean', 'wipe_clean._rich']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['wipe-clean = wipe_clean.main:_outer_cli']}

setup_kwargs = {
    'name': 'wipe-clean',
    'version': '0.3.1',
    'description': 'Clear your terminal in a ritual way. Zero dependency.',
    'long_description': '\n<!-- -------------------------------------------------------------\n\nREADME.md is auto-generated. DO NOT MODIFY THIS FILE MANUALLY.\n\n--------------------------------------------------------------- -->\n\n\n<div align="center">\n\n<a href="https://github.com/wenoptics/python-wipe-clean">\n  <img src="https://github.com/wenoptics/python-wipe-clean/blob/master/doc/logo.png?raw=true" alt="Logo" width="160">\n</a>\n\n<h1>Wipe Clean</h1>\n\nClear your terminal in a _fun_ way. Works on Windows, Linux and macOS. **0-dependency**.\n\n[![made-with-python](https://img.shields.io/badge/Made%20with-Python-1f425f.svg)](#wipe-clean)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/wipe-clean?logo=python)](#wipe-clean)\n[![Maintainability](https://api.codeclimate.com/v1/badges/ce16faa60287059ad2ed/maintainability)](https://codeclimate.com/github/wenoptics/python-wipe-clean/maintainability)\n\n[![PyPI](https://img.shields.io/pypi/v/wipe-clean?logo=pypi)](https://pypi.org/project/wipe-clean/)\n[![PyPI - Status](https://img.shields.io/pypi/status/wipe-clean)](https://pypi.org/project/wipe-clean/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dw/wipe-clean)](https://pypi.org/project/wipe-clean/)\n\n[![Linux](https://img.shields.io/badge/Linux-FCC624?logo=linux&logoColor=black)](https://github.com/wenoptics/python-wipe-clean)\n[![Mac OS](https://img.shields.io/badge/MacOS--9cf?logo=Apple&style=social)](https://github.com/wenoptics/python-wipe-clean)\n[![Windows](https://img.shields.io/badge/Windows-0078D6?logo=windows&logoColor=white)](https://github.com/wenoptics/python-wipe-clean)\n\n<p>\n<a href="#install">Install</a> &#8226; <a href="#usages">Usages</a> &#8226; <a href="#advanced-usages">Advanced usages</a> &#8226; <a href="#roadmap">Roadmap</a> &#8226; <a href="#related-projects">Related projects</a>\n</p>\n\n\n![demo](https://github.com/wenoptics/python-wipe-clean/blob/master/doc/terminal.gif?raw=true)\n\n</div>\n\n---\n\n## Install\n\nInstall with pip:\n\n```bash\npip install wipe-clean\n```\n\n> `wipe-clean` currently requires Python 3.6.1 and above. Note that Python 3.6.0 is\n not supported due to lack of `NamedTuples` typing.\n\n\n## Usages\n\nJust:\n\n```bash\nwipe-clean\n```\n\nUse `-h, --help` to show all available options\n\n```bash\nwipe-clean -h\n```\n\n## Advanced usages\n\n### 1. Use API\n\nYou can use wipe-clean inside your project.\n\n```python\nfrom wipe_clean.main import cli as wc_cli\n\nwc_cli()\n# Or with arguments\nwc_cli(\'--frame-interval=0.005\', \'--min-frame-delay=0\')\n```\n\n### 2. Customization\n\nIt\'s possible to design your own brush shape and animation.\n\n#### Example brush\n\nTo create a new brush type, implement the `Brush` interface, e.g.\n\n```python\nfrom wipe_clean.brush import Brush, ScreenPointDrawing, ScreenPoint as P\n\nclass Wipe2x2(Brush):\n    def get_points(self, x, y, angle) -> List[ScreenPointDrawing]:\n        return [\n            ScreenPointDrawing(P(x    , y    ), \'#\'),  # noqa: E202,E203\n            ScreenPointDrawing(P(x + 1, y    ), \'#\'),  # noqa: E202,E203\n            ScreenPointDrawing(P(x    , y + 1), \'#\'),  # noqa: E202,E203\n            ScreenPointDrawing(P(x + 1, y + 1), \'#\'),\n        ]\n```\n\nThis will define a brush like this:\n\n```text\n##\n##\n```\n\n#### Example path\n\nSimilarly, you can implement the `Path` interface to create a new brush path.\n\n```python\nimport math\nfrom wipe_clean.path import Path, PathPoint, ScreenPoint as P\n\nclass MySimplePath(Path):\n    def get_points(self) -> Iterable[PathPoint]:\n        return [\n            PathPoint(P(10, 10), math.radians(45)),\n            PathPoint(P(20,  5), math.radians(0)),\n            PathPoint(P(40, 20), math.radians(90)),\n        ]\n```\n\n\n## Roadmap\n\nSee [`DEVELOPMENT.md`](./DEVELOPMENT.md)\n\n\n## Related projects\n\n- [`JeanJouliaCode/wipeclean`](https://github.com/JeanJouliaCode/wipeClean) - JavaScript version\n\n  _The first brush type (`BrushWipe`) and path animations (`PathZigZag`, `PathRectEdge`) are direct ports\n  of `JeanJouliaCode/wipeclean`. Credits go to JeanJouliaCode!_\n\n- [`Textualize/rich`](https://github.com/Textualize/rich) - _An inspiring textual UI library_\n',
    'author': 'wenoptk',
    'author_email': 'wenoptics@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wenoptics/python-wipe-clean',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
